@@ -1,5 +1,6 @@
 #!/usr/bin/env python3
 
+import time
 import requests
 
 from requests import HTTPError
@@ -7,7 +8,7 @@ from requests.compat import quote
 from requests.auth import HTTPBasicAuth
 
 
-__all__ = 'Job', 'Jenkins', 'Server', 'JenkinsError'
+__all__ = 'Job', 'Jenkins', 'Server', 'JenkinsError', 'Build'
 
 
 class Job(object):
@@ -27,38 +28,30 @@ class Job(object):
         return '%s(%r)' % (cls, self.name)
 
     def _not_exist_raise(self):
-        if not self.exists():
+        if not self.exists:
             raise JenkinsError('job "%s" does not exist' % self.name)
 
-    def exists(self):
-        '''Determine if job exists.'''
-        try:
-            self.info
-            return True
-        except JenkinsError:
-            return False
-
     def delete(self):
         '''Permanently remove job.'''
-
         self._not_exist_raise()
         url = 'job/%s/doDelete' % quote(self.name)
 
-        self.server.post(url)
-        if self.exists():
+        res = self.server.post(url, throw=False)
+        if self.exists:
             raise JenkinsError('delete of job "%s" failed' % self.name)
+        return res
 
     def enable(self):
         '''Enable job.'''
         self._not_exist_raise()
         url = 'job/%s/enable' % quote(self.name)
-        self.server.post(url)
+        return self.server.post(url)
 
     def disable(self):
         '''Disable job.'''
         self._not_exist_raise()
         url = 'job/%s/disable' % quote(self.name)
-        self.server.post(url)
+        return self.server.post(url)
 
     def build(self, parameters=None, token=None):
         '''Trigger a build.'''
@@ -74,26 +67,36 @@ class Job(object):
         else:
             url = 'job/%s/build' % self.name
 
-        self.server.post(url, params=params)
+        return self.server.post(url, params=params)
 
     def reconfigure(self, newconfig):
-        '''Update config.xml of an existing job.'''
-
+        '''Update the config.xml of an existing job.'''
         self._not_exist_raise()
 
         url = 'job/%s/config.xml' % self.name
         headers = {'Content-Type': 'text/xml'}
         params = {'name': self.name}
-        res = self.server.post(url, data=newconfig, params=params, headers=headers)
+        return self.server.post(url, data=newconfig, params=params, headers=headers)
 
     @property
     def enabled(self):
         return '<disabled>false</disabled>' in self.config
 
     @property
-    def info(self):
-        '''Fetch job information.'''
+    def exists(self):
+        '''Check if job exists.'''
+        try:
+            self.info
+            return True
+        except HTTPError as e:
+            if e.response.status_code == 404:
+                return False
+            raise
+        except JenkinsError:
+            return False
 
+    @property
+    def info(self):
         url = 'job/%s/api/json?depth=0' % quote(self.name)
         err = 'job "%s" does not exist' % self.name
         return self.server.json(url, errmsg=err)
@@ -115,6 +118,23 @@ class Job(object):
     def builds(self):
         return [Build(self, i['number']) for i in self.info['builds']]
 
+    def __last_build_helper(self, path):
+        url = 'job/%s/%s/api/json' % (quote(self.name), path)
+        res = self.server.json(url)
+        return Build(self, res['number'])
+
+    @property
+    def last_build(self):
+        return self.__last_build_helper('lastBuild')
+
+    @property
+    def last_stable_build(self):
+        return self.__last_build_helper('lastStableBuild')
+
+    @property
+    def last_successful_build(self):
+        return self.__last_build_helper('lastSuccessfulBuild')
+
     @property
     def buildnumbers(self):
         return [i['number'] for i in self.info['builds']]
@@ -124,17 +144,19 @@ class Job(object):
         '''Create a new Jenkins job.'''
 
         job = cls(name, server)
-        if job.exists():
+        if job.exists:
             raise JenkinsError('job "%s" already exists' % name)
 
         headers = {'Content-Type': 'text/xml'}
         params = {'name': name}
-        res = server.post('createItem', data=configxml, params=params, headers=headers)
+        res = server.post('createItem', data=configxml, params=params, headers=headers, throw=False)
 
         if not res or res.status_code != 200:
-            raise JenkinsError('create "%s" failed' % name, url=res.url)
+            raise JenkinsError('create "%s" failed' % name)
+        else:
+            res.raise_for_status()
 
-        # if not job.exists():
+        # if not job.exists:
         #     raise JenkinsError('create "%s" failed' % name, url=res.url)
 
     @classmethod
@@ -144,17 +166,17 @@ class Job(object):
         job = cls(source, server)
         newjob = cls(dest, server)
 
-        if newjob.exists():
+        if newjob.exists:
             raise JenkinsError('job "%s" already exists' % dest)
 
-        if not job.exists():
+        if not job.exists:
             raise JenkinsError('job "%s" does not exist' % source)
 
         headers = {'Content-Type': 'text/xml'}
         params = {'name': dest, 'mode': 'copy', 'from': source}
         res = server.post('createItem', params=params, headers=headers)
 
-        if not newjob.exists():
+        if not newjob.exists:
             msg = 'could not copy job "%s" to "%s"'
             raise JenkinsError(msg % (source, dest))
 
@@ -170,13 +192,31 @@ class Build(object):
         self.job = job
         self.number = number
 
+    @property
+    def info(self):
+        '''Get information about this build.'''
+        url = 'job/%s/%d/api/json' % (self.job.name, self.number)
+        return self.job.server.json(url, 'unable to retrieve info')
+
+    @property
+    def building(self):
+        return self.info['building']
+
     def __repr__(self):
         cls = self.__class__.__name__
         return '%s(%r, %r)' % (cls, self.job, self.number)
 
     def stop(self):
         url = 'job/%s/%d/stop' % (self.job.name, self.number)
-        return self.job.post(url)
+        return self.job.server.post(url)
+
+    def wait(self, interval=1, timeout=None):
+        '''Wait for build to complete.'''
+        start = time.time()
+        while self.building:
+            time.sleep(interval)
+            if timeout and (time.time() - start) > timeout:
+                break
 
 
 class Server(object):
@@ -191,23 +231,24 @@ class Server(object):
     def urljoin(self, *args):
         return '%s%s' % (self.url, '/'.join(args))
 
-    def post(self, url, **kw):
+    def post(self, url, throw=True, **kw):
         res = requests.post(self.urljoin(url), auth=self.auth, **kw)
+        throw and res.raise_for_status()
         return res
 
-    def get(self, url, **kw):
+    def get(self, url, throw=True, **kw):
         res = requests.get(self.urljoin(url), auth=self.auth, **kw)
+        throw and res.raise_for_status()
         return res
 
-    def json(self, url, errmsg=None, **kw):
+    def json(self, url, errmsg=None, throw=True, **kw):
         url = self.urljoin(url)
         try:
-            res = requests.get(url, **kw)
+            res = requests.get(url, auth=self.auth, **kw)
+            throw and res.raise_for_status()
             if not res:
                 raise JenkinsError(errmsg)
             return res.json()
-        except HTTPError:
-            raise JenkinsError(errmsg)
         except ValueError:
             raise JenkinsError('unparsable json response')
 
@@ -238,15 +279,19 @@ class Jenkins(object):
     def jobnames(self):
         return [i['name'] for i in self.info['jobs']]
 
-    # convenience job api
+    # alternative job and build api
     def job(self, name):
         return Job(name, self.server)
 
+    def build(self, name, number):
+        name = name.name if isinstance(name, Job) else name
+        return Build(name, number)
+
     def job_info(self, name):
         return self.job(name).info
 
     def job_exists(self, name):
-        return self.job(name).exists()
+        return self.job(name).exists
 
     def job_delete(self, name):
         return self.job(name).delete()
@@ -254,12 +299,12 @@ class Jenkins(object):
     def job_enable(self, name):
         return self.job(name).enable()
 
-    def job_enabled(self, name):
-        return self.job(name).enabled
-
     def job_disable(self, name):
         return self.job(name).disable()
 
+    def job_enabled(self, name):
+        return self.job(name).enabled
+
     def job_config(self, name):
         return self.job(name).config
 
@@ -271,16 +316,46 @@ class Jenkins(object):
     def job_build(self, name, parameters=None, token=None):
         return self.job(name).build(parameters, token)
 
+    def job_builds(self, name):
+        return self.job(name).builds
+
+    def job_last_build(self, name):
+        return self.job(name).last_build
+
+    def job_last_stable_build(self, name):
+        return self.job(name).last_stable_build
+
+    def job_last_successful_build(self, name):
+        return self.job(name).last_successful_build
+
     def job_create(self, name, config):
         return Job.create(name, config, self.server)
 
     def job_copy(self, source, dest):
         return Job.copy(source, dest, self.server)
 
+    job_exists.__doc__ = Job.exists.__doc__
+    job_delete.__doc__ = Job.delete.__doc__
+    job_enable.__doc__ = Job.enable.__doc__
+    job_disable.__doc__ = Job.disable.__doc__
+    job_reconfigure.__doc__ = Job.reconfigure.__doc__
+    job_build.__doc__ = Job.build.__doc__
+    job_create.__doc__ = Job.create.__doc__
+    job_copy.__doc__ = Job.copy.__doc__
 
 class JenkinsError(Exception):
-    '''General exception type for jenkins-API-related failures.'''
+    '''Exception type for Jenkins-API related failures.'''
 
-    def __init__(self, msg, url=None, **kw):
-        super(JenkinsError, self).__init__(msg, **kw)
-        self.url = url
+    def __init__(self, msg):
+        super(JenkinsError, self).__init__(msg)
+        self.msg = msg
+
+
+## uncomment to enable http logging
+# import logging, httplib
+# httplib.HTTPConnection.debuglevel = 1
+# logging.basicConfig()
+# logging.getLogger().setLevel(logging.DEBUG)
+# requests_log = logging.getLogger('requests.packages.urllib3')
+# requests_log.setLevel(logging.DEBUG)
+# requests_log.propagate = True
